/*! \file cgi-bin.c \brief CGI-BIN Programme und Funktionen */
//***************************************************************************
//*            cgi-bin.c
//*
//*  Sat May  10 21:07:42 2008
//*  Copyright  2008  Dirk Broßwick
//*  Email: sharandac@snafu.de
//****************************************************************************/
///	\ingroup software
///	\defgroup cgibin CGI-BIN Programme und Funktionen (cgi-bin.c)
///	\code #include "cgi-bin.h" \endcode
///	\par Uebersicht
//****************************************************************************/
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */
//@{
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/io.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "apps/telnet/telnet.h"
#include "config.h"

#include "hardware/network/enc28j60.h"

#include "system/clock/clock.h"
#include "system/net/ethernet.h"
#include "system/stdout/stdout.h"
#include "system/config/config.h"
 
#ifdef __AVR_ATmega2561__
	#include "apps/mp3-streamingclient/mp3-clientserver.h"
#endif
#ifdef __AVR_ATmega644__
	#include "hardware/adc/adc.h"
	#include "hardware/gpio/gpio_out.h"
	#include "hardware/gpio/gpio_in.h"
	#include "hardware/1wire/1wire.h"
#endif

#include "httpd2.h" 
#include "httpd2_pharse.h" 
#include "cgi-bin.h"
#include "system/net/arp.h"
#include "system/net/ethernet.h"
#include "system/net/tcp.h"
#include "system/net/ip.h"
#include "system/net/ntp.h"
#include "system/softreset/softreset.h"

void cgi_stream( void * pStruct );
void cgi_dio_out( void * pStruct );
void cgi_aio( void * pStruct );
void cgi_stats( void * pStruct );
void cgi_reset( void * pStruct );
void cgi_network( void * pStruct );
void cgi_ntp( void * pStruct );
void cgi_eemem( void * pStruct );
void cgi_dio_in( void * pStruct );
void cgi_onewire(  void * pStruct );


#ifdef __AVR_ATmega644__
	#ifdef HTTPSERVER_GPIO
		const char cmd1[] PROGMEM = "dio_out.cgi";
		const char cmd8[] PROGMEM = "dio_in.cgi";
	#endif
	#ifdef HTTPSERVER_STATS
		const char cmd2[] PROGMEM = "stats.cgi";
	#endif
	#ifdef HTTPSERVER_SYSTEM
		const char cmd3[] PROGMEM = "reset.cgi";
		const char cmd7[] PROGMEM = "eemem.cgi";
	#endif
	#ifdef HTTPSERVER_NETCONFIG
		const char cmd4[] PROGMEM = "network.cgi";
	#endif
	#ifdef HTTPSERVER_ADC
		const char cmd5[] PROGMEM = "aio.cgi";
	#endif
	#ifdef NTP
		const char cmd6[] PROGMEM = "ntp.cgi";
	#endif
	#ifdef HTTPSERVER_ONEWIRE
		const char cmd10[] PROGMEM = "onewire.cgi";
	#endif
#endif

#ifdef __AVR_ATmega2561__
	const char cmd1[] PROGMEM = "stream.cgi";
	#ifdef HTTPSERVER_STATS
		const char cmd2[] PROGMEM = "stats.cgi";
	#endif
	#ifdef HTTPSERVER_SYSTEM
		const char cmd3[] PROGMEM = "reset.cgi";
	#endif
	#ifdef HTTPSERVER_NETCONFIG
		const char cmd4[] PROGMEM = "network.cgi";
	#endif
	#ifdef NTP
		const char cmd6[] PROGMEM = "ntp.cgi";
	#endif
	#ifdef HTTPSERVER_SYSTEM
		const char cmd7[] PROGMEM = "eemem.cgi";
	#endif
#endif


#ifdef __AVR_ATmega2561__
	CGIBIN cgibin[ ] = {
		{ cgi_stream, cmd1 },
	#ifdef HTTPSERVER_STATS
		{ cgi_stats, cmd2 },
	#endif
	#ifdef HTTPSERVER_SYSTEM
		{ cgi_eemem, cmd7 },		
		{ cgi_reset, cmd3 },
	#endif
	#ifdef HTTPSERVER_NETCONFIG
		{ cgi_network, cmd4 },
	#endif
	#ifdef NTP
		{ cgi_ntp, cmd6 },
	#endif
		{ NULL , NULL }
	};
#endif

#ifdef __AVR_ATmega644__
	CGIBIN cgibin[ ] = {
	#ifdef HTTPSERVER_GPIO
		{ cgi_dio_out, cmd1 },
	#endif
	#ifdef HTTPSERVER_STATS
		{ cgi_stats, cmd2 },
	#endif
	#ifdef HTTPSERVER_SYSTEM
		{ cgi_reset, cmd3 },
		{ cgi_eemem, cmd7 },		
	#endif
	#ifdef HTTPSERVER_NETCONFIG
		{ cgi_network, cmd4 },
	#endif
	#ifdef HTTPSERVER_ADC
		{ cgi_aio, cmd5 },
	#endif
	#ifdef NTP
		{ cgi_ntp, cmd6 },
	#endif
	#ifdef HTTPSERVER_GPIO
		{ cgi_dio_in, cmd8 },
	#endif
	#ifdef HTTPSERVER_ONEWIRE
		{ cgi_onewire, cmd10 },
	#endif
		{ NULL , NULL }
	};
#endif

/*------------------------------------------------------------------------------------------------------------*/
/*!\brief Das CGI-Interface. Sucht nach einen entsprechnden Eintrag und führt wenn vorhanden das passende CGI aus.
 * Als Template für das einfügen eines eigenen CGI sollte folgender Code benutzt werden. Damit das passende CGI gefunden
 * und ausgeführt werden kann muss der Name und die Funktion noch in die Struktur cgibin eingetrafeg werden. Das Schema ist
 * im Kopf der cgi-bin.c zu sehen.
 *
 * \code
 * void cgi_foobar( void * pStruct )
 * {
 *    struct HTTP_REQUEST * http_request;
 *    http_request = (struct HTTP_REQUEST *) pStruct;
 *
 *    // Hier sollte dann der eigene Code stehen, mit den obrigen Zeilen wird die Umgebung eingerichtet.
 *    // Danach stehen die Daten die übergeben werden zur Verfügung.
 *    // Eine einfache Abfrage ob z.b. eine Variable mit den Namen "foo" übergeben wurde könnte so aussehen.
 *    // Nicht zu vergessen ist, das ein kompletter HTML-Body gesendet werden muss. Alle ausgaben per printf
 *    // werden automatisch an den Client der die Abfrage gestellt hat übertragen.
 *
 *    printf_P( PSTR( "<HTML>\r\n"
 *                    "<HEAD>\r\n"));
 *
 *    if ( PharseCheckName_P ( http_request, PSTR("foo") ) )
 *    {
 *       printf_P( PSTR("Variable \"foo\" wurde gefunden, sie enthält folgenden Wert %s.\r\n"), 
				  http_request->argvalue[ PharseGetValue_P ( http_request, PSTR("foo") ) ] );   
 *    }
 *    else
 *    {
 *       printf_P( PSTR("Keine Variable \"foo\" übergeben!\r\n"));
 *    }
 *
 *    printf_P( PSTR( "</BODY>\r\n"
 *                    "</HTML>\r\n"
 *                    "\r\n"));
 * }
 * \endcode
 * \param 	pStruct		Pointer auf Struktur auf den HTTP_Request
 * \return	returnvalue	-1 Failed, 1 CGI gefunden
 */
/*------------------------------------------------------------------------------------------------------------*/
int check_cgibin( void * pStruct )
{
	struct HTTP_REQUEST * http_request;
	http_request = (struct HTTP_REQUEST *) pStruct;
	
	int i, returnvalue = -1;
	
	for( i = 0 ; i < MAX_CGI_ENTRYS ; i++ )
	{
		if ( cgibin[ i ].funktionname == NULL )
			break;
		
		if ( !strcmp_P( http_request->GET_FILE, cgibin[ i ].funktionname ) )
		{
			printf_P( PSTR(		"HTTP/1.0 200 Document follows\r\n"
								"Content-Type: text/html\r\n"
								"Keep-Alive: timeout=1, max=5\r\n"
								"Connection: close\r\n"
								"\r\n"));
			cgibin[ i ].programname( http_request );
			returnvalue = 1 ;
			break;
		}
	}

	return( returnvalue );
}	
	
#ifdef __AVR_ATmega2561__
/*------------------------------------------------------------------------------------------------------------*/
/*!\brief Das CGI-Streaminterface für zum Steuern der Streamingengine
 * \param 	pStruct	Struktur auf den HTTP_Request
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/
void cgi_stream( void * pStruct )
{
	struct HTTP_REQUEST * http_request;
	http_request = (struct HTTP_REQUEST *) pStruct;
	

	printf_P( PSTR(		"<HTML>\r\n"
						"<HEAD>\r\n"));
	if( http_request->argc != 0 && PharseCheckName_P ( http_request, PSTR("info") ) )
	{
		printf_P( PSTR(	"<meta http-equiv=\"refresh\" content=\"60; \">\r\n"));
	}
	
	printf_P( PSTR(	"<TITLE>Streaming</TITLE>\r\n"
					"</HEAD>\r\n"
					"<BODY>\r\n"));

	if ( http_request->argc == 0  )
	{
		printf_P( PSTR( "<form action=\"stream.cgi\" method=\"get\" accept-charset=\"ISO-8859-1\">"
					    "<p>Stream-URL:<input name=\"streamurl\" size=\"50\"></p>"
					    "<p><input type=\"submit\" value=\"stream starten\"></p></form>"
					   	"<form action=\"stream.cgi?stop\" method=\"get\" accept-charset=\"ISO-8859-1\">"
					    "<p><input type=\"submit\" name=\"stop\" value=\"stop\"></p></form>"
					   	"<form action=\"stream.cgi?replay\" method=\"get\" accept-charset=\"ISO-8859-1\">"
					    "<p><input type=\"submit\" name=\"replay\" value=\"replay\"></p></form>" ));
	}
	else if( PharseCheckName_P ( http_request, PSTR("streamurl") ) )
	{	
		printf_P( PSTR("<pre><p>"));
		PlayURL( http_request->argvalue[ PharseGetValue_P ( http_request, PSTR("streamurl") ) ], http_request->HTTP_SOCKET );
		printf_P( PSTR("</pre></p>"));
	}
	else if( PharseCheckName_P ( http_request, PSTR("config") ) )
	{	
		printf_P( PSTR("kommt noch!"));
	}
	else if( PharseCheckName_P ( http_request, PSTR("replay") ) )
	{	
		printf_P( PSTR("<pre><p>"));
		RePlayURL( http_request->HTTP_SOCKET );
		printf_P( PSTR("</pre></p>"));
	}
	else if( PharseCheckName_P ( http_request, PSTR("stop") ) )
	{	
		printf_P( PSTR("<pre><p>"));
		StopPlay();
		printf_P( PSTR("</pre></p>"));
	}
	else if( PharseCheckName_P ( http_request, PSTR("info") ) )
	{
		printf_P( PSTR("<pre><p>"));
		mp3clientPrintInfo ( http_request->HTTP_SOCKET , HTML );
		printf_P( PSTR("</pre></p>"));
	}
	printf_P( PSTR(		"</BODY>\r\n"
						"</HTML>\r\n"
						"\r\n"));
	
}
#endif

#ifdef HTTPSERVER_STATS
/*------------------------------------------------------------------------------------------------------------*/
/*!\brief Das CGI-Interface um Stats anzuzeigen (Uhrzeit, Tranfervolumen etc.).
 * \param 	pStruct	Struktur auf den HTTP_Request
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/
void cgi_stats( void * pStruct )
{	
	
	struct HTTP_REQUEST * http_request;
	http_request = (struct HTTP_REQUEST *) pStruct;
	
	static int VisitCounter=0;

	long time;
	
	struct TIME Time;
	// Zeit holen
	CLOCK_GetTime ( &Time );
	
	VisitCounter++;
			

	printf_P( PSTR(	"<HTML>\r\n"
					"<HEAD>\r\n"
					"<meta http-equiv=\"refresh\" content=\"600; \">\r\n"
					"<TITLE>Streaming</TITLE>\r\n"
					"</HEAD>\r\n"
					"<BODY bgcolor=\"#228B22\" text=\"#FFFFFF\">"));
			
	printf_P( PSTR( "Zeit: %02d:%02d:%02d.%02d"), Time.hh, Time.mm, Time.ss, Time.ms );
		
	printf_P( PSTR( ", Uptime: "));
	
	time = Time.uptime / ( 86400l );
	if ( time != 0 )
		if ( time == 1 ) printf_P(PSTR("%ld tag "), time );
		else printf_P(PSTR("%ld tage "), time );

	time = ( Time.uptime % ( 86400l ) ) / 3600;
	if ( time != 0 ) printf_P(PSTR("%ld std "), time );

	time = ( Time.uptime % 3600 ) / 60;
	if ( time != 0 ) printf_P(PSTR("%ld min "), time );

	time = Time.uptime % 60;
	if ( time != 0 ) printf_P(PSTR("%ld sek"), time );
	
	printf_P(PSTR(", "), time );
	
	printf_P( PSTR( "Ethernet:"));
	
	if ( ByteCounter < 1024 )
		printf_P( PSTR(" %ld Bytes "), ByteCounter);
	else if ( ByteCounter < (1024l*1024l) )
		printf_P( PSTR(" %ld kBytes "), ByteCounter / 1024l );
	else if ( ByteCounter < (1024l*1024l*1024l) )
		printf_P( PSTR(" %ld.%03ld MBytes "), ByteCounter / ( 1024l * 1024l ), ( ByteCounter % ( 1024l * 1024l ) ) / 1024l  );
	else
		printf_P( PSTR(" %ld.%03ld GBytes "), ByteCounter / ( 1024l * 1024l *1024l ), ( ByteCounter % ( 1024l * 1024l * 1024l ) ) / ( 1024l * 1024l ) );
		
	printf_P( PSTR( "in %ld Packeten ; "), PacketCounter );

	printf_P( PSTR( "Du bist der %d. Besucher auf Socket %d.\r\n"), VisitCounter, http_request->HTTP_SOCKET );

	printf_P( PSTR(	"</BODY>\r\n"
					"</HTML>\r\n"
					"\r\n"));

}
#endif

/*------------------------------------------------------------------------------------------------------------*/
/*!\brief Das CGI-Resetinterface.
 * \param 	pStruct	Struktur auf den HTTP_Request
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/
void cgi_reset( void * pStruct )
{
	struct HTTP_REQUEST * http_request;
	http_request = (struct HTTP_REQUEST *) pStruct;
	
	printf_P( PSTR(		"<HTML>\r\n"
						"<HEAD>\r\n"
						"<TITLE>Reset</TITLE>\r\n"
						"</HEAD>\r\n"
						"<BODY>\r\n"));

	if ( http_request->argc == 0 )
	{
		printf_P( PSTR(	"<form action=\"reset.cgi\" method=\"get\" accept-charset=\"ISO-8859-1\">"
						"<p><input type=\"submit\" name=\"reset\" value=\"reset\"></p></form>"
						"</BODY>\r\n"
						"</HTML>\r\n"
						"\r\n"));
	}
	else if( PharseCheckName_P( http_request, PSTR("reset") ) )
	{	
		if ( !strcmp_P( http_request->argvalue[ PharseGetValue_P( http_request, PSTR("reset") ) ] , PSTR("reset") ) )
		{
			printf_P( PSTR(	"Reset</BODY>\r\n"
							"</HTML>\r\n"
							"\r\n"));
			STDOUT_Flush();
			CloseTCPSocket( http_request->HTTP_SOCKET );
			softreset();
			while(1);
		}
	}
}

#ifdef __AVR_ATmega644__
#ifdef HTTPSERVER_ADC
/*------------------------------------------------------------------------------------------------------------*/
/*!\brief Das CGI-Interface für zum anzeigen der Analogeingänge
 * \param 	pStruct	Struktur auf den HTTP_Request
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/
void cgi_aio( void * pStruct )
{
	struct HTTP_REQUEST * http_request;
	http_request = (struct HTTP_REQUEST *) pStruct;
	
	char i;
	char gpio[10];
	int temp;

	printf_P( PSTR(	"<HTML>\r\n"
					"<HEAD>\r\n"
					"<TITLE>ADC</TITLE>\r\n"
					"</HEAD>\r\n"
					"<BODY>\r\n"));
	printf_P( PSTR(	"<form action=\"dio.cgi\">\r"
					" <table border=\"0\" cellpadding=\"5\" cellspacing=\"0\">" ) );
	for( i = 0 ; i < 8 ; i++ )
	{
		temp = ADC_mVolt( ADC_GetValue( i ) + 1 , 5000 );
		printf_P( PSTR(	"  <tr>\r"
					    "   <td align=\"right\">ADC%d =</td>"
					   	"   <td align=\"right\"> %d.%03dV</td>"
 						"  </tr>\r"), i, temp / 1000, temp % 1000 );
	}
	printf_P( PSTR(	" </table>\r"
					"</form>" ) );	
	printf_P( PSTR(	"</BODY>\r\n"
					"</HTML>\r\n"
					"\r\n"));
}
#endif

#ifdef HTTPSERVER_GPIO
/*------------------------------------------------------------------------------------------------------------*/
/*!\brief Das CGI-Interface für zum Anzeigen der Digitalen Eingänge.
 * \param 	pStruct	Struktur auf den HTTP_Request
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/
void cgi_dio_in( void * pStruct )
{
	struct HTTP_REQUEST * http_request;
	http_request = (struct HTTP_REQUEST *) pStruct;
	
	char i;
	char gpio[10];
		
	printf_P( PSTR(	"<HTML>\r\n"
					"<HEAD>\r\n"
					"<TITLE>GPIO</TITLE>\r\n"
					"</HEAD>\r\n"
					"<BODY>\r\n"));
	printf_P( PSTR(	"<form action=\"dio_out.cgi\">\r"
					" <table border=\"0\" cellpadding=\"5\" cellspacing=\"0\">" ) );

	for( i = 0 ; i < 8 ; i++ )
	{
		printf_P( PSTR(	"  <tr>\r"
					   	"   <td align=\"right\">GPIO %d =</td>" ), i);

		if ( GPIO_in_state( i ) != 0 )
		{
			
			printf_P( PSTR( "   <td>On</td>\r\n" ), i); 
		}
		else
		{
			printf_P( PSTR( "   <td>Off</td>" ), i); 
		}
		printf_P( PSTR(	"   </td>\r"
 						"  </tr>\r") );
	}
	printf_P( PSTR(	" </table>\r"
					"</form>" ) );	
	printf_P( PSTR(	"</BODY>\r\n"
					"</HTML>\r\n"
					"\r\n"));
}

/*------------------------------------------------------------------------------------------------------------*/
/*!\brief Das CGI-Interface für zum Steuern der Digitalen Ausgänge
 * \param 	pStruct	Struktur auf den HTTP_Request
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/
void cgi_dio_out( void * pStruct )
{
	struct HTTP_REQUEST * http_request;
	http_request = (struct HTTP_REQUEST *) pStruct;
	
	char i;
	char gpio[10];
	
	if ( http_request->argc != 0 )
	{
		for( i = 0 ; i < 8 ; i++ )
		{
			sprintf_P( gpio, PSTR("GPIO%d"), i );
			if ( PharseCheckName( http_request, gpio ) ) GPIO_out_set( i );
			else GPIO_out_clear( i );
		}
	}

	
	printf_P( PSTR(	"<HTML>\r\n"
					"<HEAD>\r\n"
					"<TITLE>GPIO</TITLE>\r\n"
					"</HEAD>\r\n"
					"<BODY>\r\n"));
	printf_P( PSTR(	"<form action=\"dio_out.cgi\">\r"
					" <table border=\"0\" cellpadding=\"5\" cellspacing=\"0\">" ) );
	for( i = 0 ; i < 8 ; i++ )
	{
		printf_P( PSTR(	"  <tr>\r"
					   	"   <td align=\"right\">GPIO %d</td>" ), i);

		if ( GPIO_out_state( i ) != 0 )
		{
			
			printf_P( PSTR( "   <td><input type=\"checkbox\" name=\"GPIO%d\" checked" ), i); 
		}
		else
		{
			printf_P( PSTR( "   <td><input type=\"checkbox\" name=\"GPIO%d\"" ), i); 
		}
		printf_P( PSTR(	"></td>\r"
 						"  </tr>\r") );
	}
	printf_P( PSTR(	"  <tr>\r"
					"   <td></td><td><input type=\"submit\" name=\"SUB\"></td>"
					"  </tr>\r"
				   	" </table>\r"
					"</form>" ) );	
	printf_P( PSTR(	"</BODY>\r\n"
					"</HTML>\r\n"
					"\r\n"));
}
#endif
#endif

#ifdef HTTPSERVER_NETCONFIG
/*------------------------------------------------------------------------------------------------------------*/
/*!\brief Das CGI-Interface für zum Einstellen und Anzeigen der Netzwerkeinstellungen.
 * \param 	pStruct	Struktur auf den HTTP_Request
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/
void cgi_network( void * pStruct )
{
	struct HTTP_REQUEST * http_request;
	http_request = (struct HTTP_REQUEST *) pStruct;
	
	unsigned char IP[32];

	printf_P( PSTR(		"<HTML>\r\n"
						"<HEAD>\r\n"
						"<TITLE>network</TITLE>\r\n"
						"</HEAD>\r\n"
						" <BODY>\r\n"));

	if ( http_request->argc == 0 )
	{
					   
		printf_P( PSTR( "<pre>\r\n") );
#ifdef TELNETSERVER
		TELNET_runcmdextern( "ifconfig" );
		TELNET_runcmdextern( "stats" );
#endif		
		printf_P( PSTR( "</pre>\r\n") );
	}
	else if( PharseCheckName_P( http_request, PSTR("config") ) )
	{	
		
		printf_P( PSTR(	"<form action=\"network.cgi\">\r"
					   	" <table border=\"0\" cellpadding=\"5\" cellspacing=\"0\">" ));
#ifdef DHCP
		readConfig_P( PSTR("DHCP"), IP );
  		printf_P( PSTR( "  <tr>\r"
					   	"   <td align=\"right\">DHCP aktivieren</td>"
					    "   <td><input type=\"checkbox\" name=\"DHCP\" value=\"on\" " )); 
		if ( !strcmp_P( IP, PSTR("on") ) )
						printf_P( PSTR("checked"));
		printf_P( PSTR(	"></td>\r"
  						"  </tr>\r") );
#endif
		if ( readConfig_P ( PSTR("IP"), IP ) != 1 ) iptostr ( myIP, IP );
  		printf_P( PSTR(	"  <tr>\r"
					   	"   <td align=\"right\">IP-Adresse</td>"
  						"   <td><input name=\"IP\" type=\"text\" size=\"15\" value=\"%s\" maxlength=\"15\"></td>"
  						"  </tr>\r") , IP );
		if ( readConfig_P ( PSTR("MASK"), IP ) != 1 ) iptostr ( Netmask, IP );		
  		printf_P( PSTR(	"  <tr>\r"
					   	"   <td align=\"right\">Netzwerkmaske</td>"
  						"   <td><input name=\"MASK\" type=\"text\" size=\"15\" value=\"%s\" maxlength=\"15\"></td>"
  						"  </tr>\r") , IP );
		if ( readConfig_P ( PSTR("GATE"), IP ) != 1 ) iptostr ( Gateway, IP );		
  		printf_P( PSTR(	"  <tr>\r"
					   	"   <td align=\"right\">Gateway</td>"
						"   <td><input name=\"GATE\" type=\"text\" size=\"15\" value=\"%s\" maxlength=\"15\"></td>"
  						"  </tr>\r") , IP );
#ifdef DNS
		if ( readConfig_P ( PSTR("DNS"), IP ) != 1 )  iptostr ( DNSserver, IP );		
  		printf_P( PSTR(	"  <tr>\r"
					   	"   <td align=\"right\">DNS-Server</td>"
						"   <td><input name=\"DNS\" type=\"text\" size=\"15\" value=\"%s\" maxlength=\"15\"></td>"
  						"  </tr>\r" ), IP );
#endif
		if ( readConfig_P ( PSTR("MAC"), IP ) != 1 ) mactostr( mymac, IP );
  		printf_P( PSTR(	"  <tr>\r"
					   	"   <td align=\"right\">Hardware-MAC</td>"
						"   <td><input name=\"MAC\" type=\"text\" size=\"17\" value=\"%s\" maxlength=\"17\"></td>"
  						"  </tr>\r" ) , IP );
		printf_P( PSTR(	"  <tr>\r"
   						"   <td></td><td><input type=\"submit\" value=\" Einstellung &Uuml;bernehmen \"></td>"
  						"  </tr>\r"
					   	" </table>\r"
						"</form>" ) );
	}
	else
	{
#ifdef DHCP
		if ( PharseCheckName_P( http_request, PSTR("DHCP") ) )
			changeConfig_P( PSTR("DHCP"), http_request->argvalue[ PharseGetValue_P ( http_request, PSTR("DHCP") ) ] );
		else
			changeConfig_P( PSTR("DHCP"), "on" );
#endif
		if ( PharseCheckName_P( http_request, PSTR("IP") ) )
			changeConfig_P( PSTR("IP"), http_request->argvalue[ PharseGetValue_P ( http_request, PSTR("IP") ) ] );
		if ( PharseCheckName_P( http_request, PSTR("MASK") ) )
			changeConfig_P( PSTR("MASK"), http_request->argvalue[ PharseGetValue_P ( http_request, PSTR("MASK") ) ] );
		if ( PharseCheckName_P( http_request, PSTR("GATE") ) )
			changeConfig_P( PSTR("GATE"), http_request->argvalue[ PharseGetValue_P ( http_request, PSTR("GATE") ) ] );
		#ifdef DNS
		if ( PharseCheckName_P( http_request, PSTR("DNS") ) )
			changeConfig_P( PSTR("DNS"), http_request->argvalue[ PharseGetValue_P ( http_request, PSTR("DNS") ) ] );
		#endif
		if ( PharseCheckName_P( http_request, PSTR("MAC") ) )
			changeConfig_P( PSTR("MAC"), http_request->argvalue[ PharseGetValue_P ( http_request, PSTR("MAC") ) ] );
		printf_P( PSTR("Einstellungen uebernommen!\r\n"));
			
	}
	
	printf_P( PSTR( " </BODY>\r\n"
					"</HTML>\r\n"
					"\r\n"), IP);

}
#endif

#ifdef NTP
/*------------------------------------------------------------------------------------------------------------*/
/*!\brief Das CGI-Interface für zum Einstellen und Anzeigen für den NTP-Service.
 * \param 	pStruct	Struktur auf den HTTP_Request
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/
void cgi_ntp( void * pStruct )
{
	struct HTTP_REQUEST * http_request;
	http_request = (struct HTTP_REQUEST *) pStruct;
	
	char NTPSERVER[32];
	char UTCZONE[4];

	printf_P( PSTR(		"<HTML>\r\n"
						"<HEAD>\r\n"
						"<TITLE>SNTP</TITLE>\r\n"
						"</HEAD>\r\n"
						" <BODY>\r\n"));

	if ( http_request->argc == 0 )
	{
		readConfig_P( PSTR("NTP"), NTPSERVER );
		printf_P( PSTR(	"<form action=\"ntp.cgi\">\r"
					   	" <table border=\"0\" cellpadding=\"5\" cellspacing=\"0\">"
  						"  <tr>\r"
					   	"   <td align=\"right\">SNTP aktivieren</td>"
					    "   <td><input type=\"checkbox\" name=\"NTP\" value=\"on\" " )); 
		if ( !strcmp_P( NTPSERVER, PSTR("on") ) )
						printf_P( PSTR("checked"));
		printf_P( PSTR(	"></td>\r"
  						"  </tr>\r") );
		
		if( checkConfigName_P( PSTR("NTP") ) != -1 )
			readConfig_P ( PSTR("NTPSERVER"), NTPSERVER );
  		else
			strcpy_P ( NTPSERVER, PSTR("time.fu-berlin.de"));
			
		if( checkConfigName_P( PSTR("UTCZONE") ) != -1 )
			readConfig_P ( PSTR("UTCZONE"), UTCZONE );
  		else
			strcpy_P ( UTCZONE, PSTR("0"));

		printf_P( PSTR(	"  <tr>\r"
					   	"   <td align=\"right\">SNTP-Server</td>"
  						"   <td><input name=\"NTPSERVER\" type=\"text\" size=\"25\" value=\"%s\" maxlength=\"25\"></td>"
  						"  </tr>\r"
						"  <tr>\r"
					   	"   <td align=\"right\">Zeitzone</td>"
						"   <td><input name=\"UTCZONE\" type=\"text\" size=\"3\" value=\"%s\" maxlength=\"3\"></td>"
  						"  </tr>\r"
 						"  <tr>\r"
   						"   <td></td><td><input type=\"submit\" value=\" Einstellung &Uuml;bernehmen \"></td>"
  						"  </tr>\r"
					   	" </table>\r"
						"</form>" ) , NTPSERVER, UTCZONE );
	}
	else
	{
		if ( PharseCheckName_P( http_request, PSTR("NTP") ) )
			changeConfig_P( PSTR("NTP"), http_request->argvalue[ PharseGetValue_P ( http_request, PSTR("NTP") ) ] );
		else
			changeConfig_P( PSTR("NTP"), "off" );	
		if ( PharseCheckName_P( http_request, PSTR("NTPSERVER") ) )
			changeConfig_P( PSTR("NTPSERVER"), http_request->argvalue[ PharseGetValue_P ( http_request, PSTR("NTPSERVER") ) ] );
		if ( PharseCheckName_P( http_request, PSTR("UTCZONE") ) )
			changeConfig_P( PSTR("UTCZONE"), http_request->argvalue[ PharseGetValue_P ( http_request, PSTR("UTCZONE") ) ] );
		
		printf_P( PSTR("Einstellungen uebernommen!\r\n"));

	}
	
	printf_P( PSTR( " </BODY>\r\n"
					"</HTML>\r\n"
					"\r\n"));

}
#endif

/*------------------------------------------------------------------------------------------------------------*/
/*!\brief Das CGI-Interface für zum Anzeigen der Konfigurationsdaten im EEProm. 
 * \param 	pStruct	Struktur auf den HTTP_Request
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/
void cgi_eemem( void * pStruct )
{
	struct HTTP_REQUEST * http_request;
	http_request = (struct HTTP_REQUEST *) pStruct;
	
	printf_P( PSTR(	"<HTML>\r\n"
					"<HEAD>\r\n"
					"<TITLE>EEmem</TITLE>\r\n"
					"</HEAD>\r\n"
					" <BODY>\r\n"
					"  <pre>\r\n")); 

	PrintConfig ();

	printf_P( PSTR( "  </pre>\r\n"
					" </BODY>\r\n"
					"</HTML>\r\n"
					"\r\n"));

}


/*------------------------------------------------------------------------------------------------------------*/
/*!\brief Das CGI-Interface für zum Anzeigen der Konfigurationsdaten im EEProm. 
 * \param 	pStruct	Struktur auf den HTTP_Request
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/
void cgi_onewire(  void * pStruct )
{
    struct HTTP_REQUEST * http_request;
    http_request = (struct HTTP_REQUEST *) pStruct;
 
	char id[8], diff;
	char i;
	unsigned int TEMP;

	printf_P( PSTR(	"<HTML>\r\n"
					"<HEAD>\r\n"
					"<TITLE>1-Wire</TITLE>\r\n"
					"</HEAD>\r\n"
					" <BODY>\r\n"
					"  <pre>\r\n"
					"Folgende Geraete wurden gefunden:\r\n")); 

	for( diff = SEARCH_FIRST; diff != LAST_DEVICE; )
	{
		cli();
		diff = ONEWIRE_searchrom( diff, id );
		sei();
				
		if( diff == PRESENCE_ERR )
		{
			printf_P( PSTR("No Sensor found\r\n" ));
			break;
		}
		if( diff == DATA_ERR )
		{
			printf_P( PSTR( "Bus Error\r\n" ));
			break;
		}

		printf_P( PSTR( "ID: " ) );
  		for( i = 0; i < 7; i++ )
		{
			printf_P( PSTR("%02X:"), id[i] );
		}
		printf_P( PSTR("%02X ("), id[ i ] );
		printf_P( ONEWIRE_getfamilycode2string ( id[0] ) );
		if( id[0] == 0x28 || id[0] == 0x22 || id[0] == 0x10 )
			{
				ONEWIRE_command( CONVERT_T, id );
				ONEWIRE_ParasitepowerOn( );
				CLOCK_delay( 750 );
				ONEWIRE_command( READ , id );
				TEMP = ONEWIRE_readbyte( );
				TEMP |= ( ONEWIRE_readbyte( ) << 8 );
				if( id[0] == 0x10 )			// 9 -> 12 bit
    				TEMP <<= 3;
				printf_P( PSTR(" (%d.%01dC)"), TEMP >> 4, (TEMP << 12) / 6553 ); // 0.1øC
			}
		printf_P( PSTR(")\r\n") );
	}
				 
	printf_P( PSTR( "  </pre>\r\n"
					" </BODY>\r\n"
					"</HTML>\r\n"
					"\r\n"));
}
//@}
