/*! \file enc28j60.c \brief Microchip ENC28J60 Ethernet Interface Driver. */
//*****************************************************************************
//
// File Name	: 'enc28j60.c'
// Title		: Microchip ENC28J60 Ethernet Interface Driver
// Author		: Pascal Stang (c)2005
// Created		: 9/22/2005
// Revised		: 9/22/2005
// Version		: 0.1
// Target MCU	: Atmel AVR series
// Editor Tabs	: 4
//
// Description	: This driver provides initialization and transmit/receive
//	functions for the Microchip ENC28J60 10Mb Ethernet Controller and PHY.
// This chip is novel in that it is a full MAC+PHY interface all in a 28-pin
// chip, using an SPI interface to the host processor.
//
//*****************************************************************************
#include <avr/io.h>
#include <avr/interrupt.h>
#include "enc28j60.h"

// #include "hardware/spi/spi_1.h"

#include "hardware/spi/spi_core.h"

#include "system/clock/clock.h"
#include "system/math/math.h"

#include "hardware/led/led_core.h"

unsigned char Enc28j60Bank;
unsigned int NextPacketPtr;
unsigned char REVID;

#ifdef __AVR_ATmega2561__
	#define SPIBUS 1

	#define ENC28J60_CONTROL_PORT	PORTE
	#define ENC28J60_CONTROL_DDR	DDRE
	#define ENC28J60_CONTROL_CS		PE3
#endif

#ifdef __AVR_ATmega644__
	#define SPIBUS 0

	#define ENC28J60_CONTROL_PORT	PORTB
	#define ENC28J60_CONTROL_DDR	DDRB
	#define ENC28J60_CONTROL_CS		PB4
#endif

//*********************************************************************************************************
//
// Setzt die MAC-Adressse im Enc28j60
//
//*********************************************************************************************************
void nicSetMacAddress( unsigned char * MAC)
{
	// write MAC address
	// NOTE: MAC address in ENC28J60 is byte-backward
  	enc28j60Write(MAADR5, MAC[ 0 ] );
	enc28j60Write(MAADR4, MAC[ 1 ] );
	enc28j60Write(MAADR3, MAC[ 2 ] );
	enc28j60Write(MAADR2, MAC[ 3 ] );
	enc28j60Write(MAADR1, MAC[ 4 ] );
	enc28j60Write(MAADR0, MAC[ 5 ] );
}

//*********************************************************************************************************
//
// Sende Rad Command
//
//*********************************************************************************************************
 unsigned char enc28j60ReadOp( unsigned char op,  unsigned char address)
{
	unsigned char data;
	// CS aktive setzen
	ENC28J60_CONTROL_PORT &= ~(1<<ENC28J60_CONTROL_CS);
	// lesecomando schreiben
	data = SPI_ReadWrite( SPIBUS, op | (address & ADDR_MASK) );
	// dummy senden um ergebnis zu erhalten
	data = SPI_ReadWrite( SPIBUS, 0x00 );
	// dummy read machen
	if ( address & 0x80 )
		data = SPI_ReadWrite( SPIBUS, 0x00 );
	// CS wieder freigeben
	ENC28J60_CONTROL_PORT |= (1<<ENC28J60_CONTROL_CS);
	
	return data;
}

//*********************************************************************************************************
//
// Sende Write Command
//
//*********************************************************************************************************
void enc28j60WriteOp( unsigned char op, unsigned char address,  unsigned char data)
{
	// CS aktive setzen
	ENC28J60_CONTROL_PORT &= ~(1<<ENC28J60_CONTROL_CS);
	// schreibcomando senden
	SPI_ReadWrite( SPIBUS, op | (address & ADDR_MASK) );
	// daten senden
	SPI_ReadWrite( SPIBUS, data );
	// CS wieder freigeben
	ENC28J60_CONTROL_PORT |= (1<<ENC28J60_CONTROL_CS);
}

//*********************************************************************************************************
//
// Buffer einlesen
//
//*********************************************************************************************************
void enc28j60ReadBuffer(unsigned int len, unsigned char * data)
{
	// assert CS
	ENC28J60_CONTROL_PORT &= ~(1<<ENC28J60_CONTROL_CS);
	
	// issue read command
	SPI_ReadWrite( SPIBUS, ENC28J60_READ_BUF_MEM );

	SPI_ReadBlock( SPIBUS, data, len);
/*	while(len--)
	{
		// read data
		*data++ = SPI1_ReadWrite( 0x00 );
	} */

	// release CS

	ENC28J60_CONTROL_PORT |= (1<<ENC28J60_CONTROL_CS);
}

//*********************************************************************************************************
//
// Buffer schreiben
//
//*********************************************************************************************************
void enc28j60WriteBuffer(unsigned int len, unsigned char * data)
{
	// assert CS
	ENC28J60_CONTROL_PORT &= ~(1<<ENC28J60_CONTROL_CS);

	// issue write command
	SPI_ReadWrite( SPIBUS, ENC28J60_WRITE_BUF_MEM );

//	SPI1_FastMem2Write( data, len );
	while(len--)
	{
		// write data
		SPI_ReadWrite( SPIBUS, *data++ );
	}
	// release CS
	ENC28J60_CONTROL_PORT |= (1<<ENC28J60_CONTROL_CS);
}

//*********************************************************************************************************
//
// 
//
//*********************************************************************************************************
void enc28j60SetBank(unsigned char address)
{
	// set the bank (if needed)
	if((address & BANK_MASK) != Enc28j60Bank)
	{
		// set the bank
		enc28j60WriteOp(ENC28J60_BIT_FIELD_CLR, ECON1, (ECON1_BSEL1|ECON1_BSEL0));
		enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, (address & BANK_MASK)>>5);
		Enc28j60Bank = (address & BANK_MASK);
	}
}

//*********************************************************************************************************
//
// 
//
//*********************************************************************************************************
unsigned char enc28j60Read(unsigned char address)
{
	// set the bank
	enc28j60SetBank(address);
	// do the read
	return enc28j60ReadOp(ENC28J60_READ_CTRL_REG, address);
}

//*********************************************************************************************************
//
// 
//
//*********************************************************************************************************
void enc28j60Write(unsigned char address, unsigned char data)
{
	// set the bank
	enc28j60SetBank(address);
	// do the write
	enc28j60WriteOp(ENC28J60_WRITE_CTRL_REG, address, data);
}

//*********************************************************************************************************
//
// 
//
//*********************************************************************************************************
unsigned int enc28j60PhyRead(unsigned char address)
{
     unsigned int data;
 
     // Set the right address and start the register read operation
     enc28j60Write(MIREGADR, address);
     enc28j60Write(MICMD, MICMD_MIIRD);
 
    // wait until the PHY read completes
     while(enc28j60Read(MISTAT) & MISTAT_BUSY);
 
     // quit reading
     enc28j60Write(MICMD, 0x00);
     
     // get data value
     data  = enc28j60Read(MIRDL);
     data |= enc28j60Read(MIRDH)<<8;
     // return the data
     return data;
}

//*********************************************************************************************************
//
// 
//
//*********************************************************************************************************
void enc28j60PhyWrite(unsigned char address, unsigned int data)
{
	// set the PHY register address
	enc28j60Write(MIREGADR, address);

	// write the PHY data
	enc28j60Write(MIWRL, data);
	enc28j60Write(MIWRH, data>>8);

	// wait until the PHY write completes
	while(enc28j60Read(MISTAT) & MISTAT_BUSY);
}

//*********************************************************************************************************
//
// Initialiesiert den ENC28J60
//
//*********************************************************************************************************
void enc28j60Init(void)
{
	// initialize I/O

	// schau mal ob SPI schon laufen tut, wenn starten
	// if ( SPI_GetInitState() == SPI_NOT_INIT ) 
	// delay(10);
	
	unsigned char mac[6] = { ENC28J60_MAC0,ENC28J60_MAC1,ENC28J60_MAC2,ENC28J60_MAC3,ENC28J60_MAC4,ENC28J60_MAC5 };
	
	SPI_init( SPIBUS );
			
	ENC28J60_CONTROL_DDR |= 1<<ENC28J60_CONTROL_CS;
	ENC28J60_CONTROL_PORT |= 1<<ENC28J60_CONTROL_CS;
	
    enc28j60WriteOp(ENC28J60_SOFT_RESET, 0, ENC28J60_SOFT_RESET);           // perform system reset

    while(!(enc28j60Read(ESTAT) & ESTAT_CLKRDY));                           // check CLKRDY bit to see if reset is complete
    
    // do bank 0 stuff
    NextPacketPtr = RXSTART_INIT;                                           // initialize receive buffer
    enc28j60Write(ERXSTL, RXSTART_INIT&0xFF);                               // 16-bit transfers, must write low byte first
    enc28j60Write(ERXSTH, RXSTART_INIT>>8);                                 // set receive buffer start address
    enc28j60Write(ERXRDPTL, RXSTART_INIT&0xFF);                             // set receive pointer address
    enc28j60Write(ERXRDPTH, RXSTART_INIT>>8);                               //
    enc28j60Write(ERXNDL, RXSTOP_INIT&0xFF);                                // ERXND defaults to 0x1FFF (end of ram)
    enc28j60Write(ERXNDH, RXSTOP_INIT>>8);                                  // set receive buffer end
    enc28j60Write(ETXSTL, TXSTART_INIT&0xFF);                               // set transmit buffer start
    enc28j60Write(ETXSTH, TXSTART_INIT>>8);                                 // ETXST defaults to 0x0000 (beginnging of ram)
    
    // do bank 2 stuff    
    enc28j60Write(MACON1, MACON1_MARXEN|MACON1_TXPAUS|MACON1_RXPAUS);       // enable MAC receive
    enc28j60Write(MACON2, 0x00);                                            // bring MAC out of reset
    enc28j60WriteOp(ENC28J60_BIT_FIELD_SET                                  //
        , MACON3, MACON3_PADCFG0|MACON3_TXCRCEN|MACON3_FRMLNEN);            // enable automatic padding and CRC operations
    enc28j60Write(MAIPGL, 0x12);                                            // set inter-frame gap (non-back-to-back)
    enc28j60Write(MAIPGH, 0x0C);                                            //
    enc28j60Write(MABBIPG, 0x12);                                           // set inter-frame gap (back-to-back)
    enc28j60Write(MAMXFLL, MAX_FRAMELEN&0xFF);                              // Set the maximum packet size which the controller will accept
    enc28j60Write(MAMXFLH, MAX_FRAMELEN>>8);                                //
    
	nicSetMacAddress( mac );

	enc28j60Write(ECOCON, 0x00);                                            // ENC Clock out disabled
 
    enc28j60Write(ERXFCON, ERXFCON_BCEN|ERXFCON_MCEN|ERXFCON_UCEN);         // receive broadcast, multicast and unicast packets
 
    //do Phy Register stuff
    enc28j60PhyWrite(PHCON2, PHCON2_HDLDIS);                                // no loopback of transmitted frames
    enc28j60SetBank(ECON1);                                                 // switch to bank 0
    enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, EIE, EIE_INTIE|EIE_PKTIE);      // enable interrutps
    enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, ECON1_RXEN);             // enable packet reception
	
}

//*********************************************************************************************************
//
// Sendet ein Packet
//
//*********************************************************************************************************
void enc28j60PacketSend(unsigned int len, unsigned char* packet)
{
    unsigned char tsv_byte[7];                                              // Array für Transmit Status Vector
    unsigned char rtx_count = 0;                                            // retransmit counter
    unsigned int count = 0;                                                 //

	for (count = 0; count < 50; count++)                                    //
    {                                                                       //
        if( !(enc28j60Read(ECON1) & ECON1_TXRTS) )                          //
        {                                                                   //
            break;                                                          //
        }                                                                   //
    }                                                                       //
    count = 0;                                                              //
    enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, ECON1_TXRST);            // Reset transmit logic if a TX Error has previously occured
    enc28j60WriteOp(ENC28J60_BIT_FIELD_CLR, ECON1, ECON1_TXRST);            // This is a silicon errata workaround
    enc28j60WriteOp(ENC28J60_BIT_FIELD_CLR, EIR, EIR_TXERIF | EIR_TXIF);    //
    enc28j60Write(EWRPTL, TXSTART_INIT&0xFF);                               // Set the write pointer to start of transmit buffer area
    enc28j60Write(EWRPTH, TXSTART_INIT>>8);                                 //
    enc28j60Write(ETXNDL, (TXSTART_INIT+len)&0xFF);                         // Set the TXND pointer to correspond to the packet size given
    enc28j60Write(ETXNDH, (TXSTART_INIT+len)>>8);                           //
    enc28j60WriteOp(ENC28J60_WRITE_BUF_MEM, 0, 0x00);                       // write per-packet control byte
    enc28j60WriteBuffer(len, packet);                                       // copy the packet into the transmit buffer
    enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, ECON1_TXRTS);            // send the contents of the transmit buffer onto the network
    if (enc28j60Read(EREVID) == 0x05 || enc28j60Read(EREVID) ==0x06)        //
    {                                                                       //
        while(!(enc28j60Read(EIR) & (EIR_TXERIF | EIR_TXIF)) && (++count < 1000));//
        if ( (enc28j60Read(EIR) & EIR_TXERIF) || (count >= 1000) )          //
        {                                                                   //
            enc28j60WriteOp(ENC28J60_BIT_FIELD_CLR, ECON1, ECON1_TXRTS);    // Cancel the previous transmission if it has become stuck set
            enc28j60Write(ERDPTL, (TXSTART_INIT+len+1)&0xFF);               // Setzt den Lese Pointer auf den anfang des Transmit Status Vector
            enc28j60Write(ERDPTH, (TXSTART_INIT+len+1)>>8);                 //
            enc28j60ReadBuffer(7, tsv_byte);                                // Liest den Transmit Status Vector
            for(rtx_count = 0; rtx_count < 16; rtx_count++)                 //
            {                                                               //
                if(enc28j60Read(EIR) & EIR_TXERIF && tsv_byte[3] & tx_latecol)//
                {                                                           //
                    enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, ECON1_TXRST);// Reset transmit logic if a TX Error has previously occured
                    enc28j60WriteOp(ENC28J60_BIT_FIELD_CLR, ECON1, ECON1_TXRST);// This is a silicon errata workaround
                    enc28j60WriteOp(ENC28J60_BIT_FIELD_CLR, EIR, EIR_TXERIF | EIR_TXIF);//
                    enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, ECON1_TXRTS);// send the contents of the transmit buffer onto the network
                    while(!(enc28j60Read(EIR) & (EIR_TXERIF | EIR_TXIF)));  //
                    enc28j60WriteOp(ENC28J60_BIT_FIELD_CLR, ECON1, ECON1_TXRTS);// Cancel the previous transmission if it has become stuck set
                    enc28j60Write(ERDPTL, (TXSTART_INIT+len+1)&0xFF);       // Setzt den Lese Pointer auf den anfang des Transmit Status Vector
                    enc28j60Write(ERDPTH, (TXSTART_INIT+len+1)>>8);         //
                    enc28j60ReadBuffer(7, tsv_byte);                        // Liest den Transmit Status Vector
                }                                                           //
                else                                                        //
                {                                                           //
                    break;                                                  //
                }                                                           //
            }                                                               //
        }                                                                   //
    }                                                                       //
 }

//*********************************************************************************************************
//
// Hole ein empfangendes Packet
//
//*********************************************************************************************************
unsigned int enc28j60PacketReceive(unsigned int maxlen, unsigned char* packet)
{
    unsigned int rsv/*, rs,re*/;                                            //
    unsigned int len;                                                       //
    if( !(enc28j60Read(EIR) & EIR_PKTIF) )                                  // prüft ob Receive Int Flag gesetzt ist 
    if (enc28j60Read(EPKTCNT) == 0)                                         // püft ob der Paket counter = 0 ist
    return 0;                                                               // wenn Int Flag nicht gesetzt oder Paket counter 0 ist dann länge 0 zurückgeben
    enc28j60Write(ERDPTL, (NextPacketPtr));                                 // Setzt den read pointer auf die Adresse die im NextPacketPtr steht
    enc28j60Write(ERDPTH, (NextPacketPtr)>>8);                              // Beim ersten durchlauf ist NextPacketPtr = RXSTART_INIT 0x0000
    NextPacketPtr   = enc28j60ReadOp(ENC28J60_READ_BUF_MEM, 0);             // Das erste Word im ReadBuffer ist die Adresse des nächsten Paketes
    NextPacketPtr  |= enc28j60ReadOp(ENC28J60_READ_BUF_MEM, 0)<<8;          // 
    len             = enc28j60ReadOp(ENC28J60_READ_BUF_MEM, 0);             // Das zweite Word im Buffer ist die Framelänge des aktuellen Paketes
    len            |= enc28j60ReadOp(ENC28J60_READ_BUF_MEM, 0)<<8;          // setzt sich zusammen aus: dest. adr, source adr., type/lenght,data,padding,crc
    rsv             = enc28j60ReadOp(ENC28J60_READ_BUF_MEM, 0);             // Das dritte Word im Buffer ist der receive status vector
    rsv            |= enc28j60ReadOp(ENC28J60_READ_BUF_MEM, 0)<<8;          // 

    if( NextPacketPtr > RXSTOP_INIT ||                                      // Validate the data returned from the ENC28J60.  Random data corruption,
        (rsv & rsv_Zero) ||                                                 // such as if a single SPI bit error occurs while communicating or a 
        (rsv & rsv_CRCError) ||                                             // momentary power glitch could cause this to occur in rare circumstances.
        (len > maxlen) ||                                                   //
        (!(rsv & rsv_ReceiveOk)))                                           //
    {                                                                       //
        NextPacketPtr = 0;
        enc28j60Init();
        return 0;
    }                                                                       //
    enc28j60ReadBuffer(len, packet);                                        // Kopiert das Paket aus dem Receive Buffer in den uC buffer
                                                                            // an implementation of Errata B1 Section #13
    //ERRATA says we need to check packet pointer:
    if ((NextPacketPtr - 1 < RXSTART_INIT) || (NextPacketPtr - 1 > RXSTOP_INIT))
    {
        enc28j60Write(ERXRDPTL, (RXSTOP_INIT&0xFF));                         // RX Read Ptr = Adr aus NextPacketPtr -1
        enc28j60Write(ERXRDPTH, (RXSTOP_INIT)>>8);                      //
    }
    else
    {
        enc28j60Write(ERXRDPTL, (NextPacketPtr-1));                         // RX Read Ptr = Adr aus NextPacketPtr -1
        enc28j60Write(ERXRDPTH, (NextPacketPtr-1)>>8);                      //
    }

    enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON2, ECON2_PKTDEC);           // decrement the packet counter indicate we are done with this packet
    return (len-4);
}

// ***********************************************************************************************************************

// Netzwerk Link prüfen **************************************************************************************************
unsigned char link_check(void)                                              //
{                                                                           //
    if(enc28j60PhyRead(0x01) & PHSTAT1_LLSTAT)                              // Prüfe ob Netzwerk Link vorhanden
    {                                                                       //
        return ( LINK_OK );                                                 //
    }                                                                       //
    else                                                                    // Wenn nicht dann
    {                                                                       //
        return ( LINK_DOWN );                                               //
    }                                                                       //
}
